from soda.sodacl.missing_and_valid_cfg import MissingAndValidCfg


class ColumnConfigurationsCfg(MissingAndValidCfg):
    """
    Default missing and validity configurations for a column.
    """

    def __init__(self, column_name: str):
        super().__init__()
        self.column_name: str = column_name



from __future__ import annotations

import numbers
from datetime import date, datetime

from soda.common.logs import Logs


class AttributeHandler:
    def __init__(self, logs: Logs) -> None:
        self.logs = logs

    def validate(self, attributes: dict[str, any], schema: list[dict]) -> tuple[dict[str, any]]:
        """Validates given attributes against given schema. Returns a dict with valid and a dict with invalid attributes."""
        valid = {}
        invalid = {}
        for k, v in attributes.items():
            is_valid = self.validate_attribute(k, v, schema)
            if is_valid:
                valid[k] = v
            else:
                invalid[k] = v

        return valid, invalid

    def validate_attribute(self, key: str, value: any, schema: list[dict]) -> bool:
        is_valid = False
        schema_type = None

        for schema_item in schema:
            if key == schema_item["name"]:
                schema_type = schema_item["type"]
                validity_method = f"is_valid_{schema_type.lower()}"

                try:
                    is_valid = getattr(self, validity_method)(key, value, schema_item)
                except AttributeError:
                    self.logs.error(f"Unsupported attribute type '{schema_type}'.")

        if not schema_type:
            self.logs.error(f"Soda Cloud does not recognize '{key}' attribute name.")

        return is_valid

    def format_attribute(self, value: any):
        # Introduce formatting methods similar to validation methods if this gets more complex.
        if isinstance(value, date):
            value = datetime.combine(value, datetime.min.time())

        if isinstance(value, datetime):
            if value.tzinfo is None:
                value = value.replace(tzinfo=datetime.now().astimezone().tzinfo)
            return value.isoformat()

        # Numeric attributes need to be sent to Cloud as strings.
        if isinstance(value, numbers.Number):
            value = str(value)

        return value

    def is_valid_checkbox(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if value not in schema["allowedValues"]:
            self._log_unrecognized_value(key, value, schema["allowedValues"])
            return False

        return True

    def is_valid_datetime(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if isinstance(value, datetime) or isinstance(value, date):
            return True

        try:
            datetime.fromisoformat(value)
            return True
        except (TypeError, ValueError):
            self.logs.error(f"Soda Cloud expects an ISO formatted date or datetime value for the '{key}' attribute.")
            return False

    def is_valid_multiselect(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if isinstance(value, list):
            invalid_values = []
            for v in value:
                if v not in schema["allowedValues"]:
                    invalid_values.append(v)

            if invalid_values:
                self._log_unrecognized_value(key, invalid_values, schema["allowedValues"])
                return False

            return True

        self.logs.error(
            f"Soda Cloud expects '{key}' to be a list. Refer to a list of valid '{key}' values in your Soda Cloud organization configuration."
        )
        return False

    def is_valid_number(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if not isinstance(value, int) and not isinstance(value, float):
            self._log_unrecognized_type(key, value, type(value).__name__, ["int", "float"])
            return False

        return True

    def is_valid_singleselect(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if value not in schema["allowedValues"]:
            self._log_unrecognized_value(key, value, schema["allowedValues"])
            return False

        return True

    def is_valid_text(self, key: str, value: any, schema: dict[str, any]) -> bool:
        if not isinstance(value, str):
            self._log_unrecognized_type(key, value, type(value).__name__, ["string"])
            return False
        return True

    def _log_unrecognized_value(self, key: str, value: any, allowed_values: list[any]) -> None:
        self.logs.error(
            f"Soda Cloud does not recognize '{key}': '{value}' attribute value. Valid attribute value(s): {allowed_values}."
        )

    def _log_unrecognized_type(self, key: str, value: any, type: str, expected_types: list(str)) -> None:
        self.logs.error(
            f"Soda Cloud does not recognize '{type}' type of attribute '{key}'. It expects the following type(s): {expected_types}"
        )



from soda.execution.query.query import Query
from soda.execution.query.sample_query import SampleQuery


class DuplicatesQuery(Query):
    def __init__(self, partition: "Partition", metric: "Metric"):
        super().__init__(
            data_source_scan=partition.data_source_scan,
            table=partition.table,
            partition=partition,
            column=metric.column,
            unqualified_query_name=f"duplicate_count",
        )
        self.metric = metric

        self.samples_limit = self.metric.samples_limit

        values_filter_clauses = [f"{column_name} IS NOT NULL" for column_name in self.metric.metric_args]
        partition_filter = self.partition.sql_partition_filter
        if partition_filter:
            scan = self.data_source_scan.scan
            resolved_partition_filter = scan.jinja_resolve(definition=partition_filter)
            values_filter_clauses.append(resolved_partition_filter)

        if metric.filter:
            values_filter_clauses.append(metric.filter)

        values_filter = " \n  AND ".join(values_filter_clauses)

        column_names = ", ".join(self.metric.metric_args)

        # This does not respect the exclude_columns config because removing any of the excluded columns here would
        # effectively change the definition of the check. Let all columns through and samples will not be collected
        # if excluded columns are present (see "gatekeeper" in Query).
        # The only way exclude columns are taken into consideration is for building up the list of columns to be
        # selected from the frequencies CTE in the main query. If no exclude columns is present, it is safe to use
        # '*', otherwise use a specific list of columns. This is a workaround for bare-bones complex types support
        # by avoiding listing complex types which have special characters in the main query as that would require
        # special handling per warehouse type like quotes.
        table_name = self.partition.table.qualified_table_name
        exclude_patterns = self.data_source_scan.data_source.get_exclude_column_patterns_for_table(table_name)
        data_source = self.data_source_scan.data_source
        jinja_resolve = self.data_source_scan.scan.jinja_resolve

        self.sql = jinja_resolve(
            data_source.sql_get_duplicates_count(
                column_names,
                table_name,
                values_filter,
            )
        )

        self.failed_rows_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                table_name,
                values_filter,
                self.samples_limit,
                exclude_patterns=exclude_patterns,
            )
        )
        self.failing_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                table_name,
                values_filter,
                None,
                exclude_patterns=exclude_patterns,
            )
        )

        self.passing_sql = jinja_resolve(
            data_source.sql_get_duplicates(
                column_names,
                self.partition.table.qualified_table_name,
                values_filter,
                None,
                invert_condition=True,
                exclude_patterns=exclude_patterns,
            )
        )

        self.failing_rows_sql_aggregated = jinja_resolve(
            data_source.sql_get_duplicates_aggregated(
                column_names,
                self.partition.table.qualified_table_name,
                values_filter,
                self.samples_limit,
                invert_condition=False,
                exclude_patterns=exclude_patterns,
            )
        )

    def execute(self):
        self.fetchone()
        duplicates_count = self.row[0]
        self.metric.set_value(duplicates_count)

        if duplicates_count and self.samples_limit > 0:
            # TODO: Sample Query execute implicitly stores the failed rows file reference in the passed on metric.
            sample_query = SampleQuery(
                self.data_source_scan,
                self.metric,
                "failed_rows",
                self.failed_rows_sql,
            )
            sample_query.execute()

        # TODO: This should be a second failed rows file, refactor failed rows to support multiple files.
        if self.failing_rows_sql_aggregated and self.samples_limit > 0:
            aggregate_sample_query = Query(
                self.data_source_scan,
                self.partition.table,
                self.partition,
                unqualified_query_name=f"duplicate_count[{'-'.join(self.metric.metric_args)}].failed_rows.aggregated",
                sql=self.failing_rows_sql_aggregated,
                samples_limit=self.samples_limit,
            )
            aggregate_sample_query.execute()
            self.aggregated_failed_rows_data = aggregate_sample_query.rows



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone

from soda.common.logs import Logs
from soda.sampler.sample import Sample


@dataclass
class SampleContext:
    sample: Sample
    sample_name: str
    query: str
    data_source: DataSource
    partition: Partition | None
    column: Column | None
    scan: Scan
    logs: Logs
    samples_limit: int | None
    passing_sql: str | None
    check_name: str | None

    def get_scan_folder_name(self):
        parts = [
            self.scan._scan_definition_name,
            self.scan._data_timestamp.strftime("%Y%m%d%H%M%S"),
            datetime.now(tz=timezone.utc).strftime("%Y%m%d%H%M%S"),
        ]
        return "_".join([part for part in parts if part])

    def get_sample_file_name(self):
        parts = [
            self.partition.table.table_name if self.partition else None,
            self.partition.partition_name if self.partition else None,
            self.column.column_name if self.column else None,
            self.sample_name,
        ]
        return "_".join([part for part in parts if part])



from typing import Dict

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.metric import Metric
from soda.execution.metric.user_defined_failed_rows_metric import (
    UserDefinedFailedRowsMetric,
)

KEY_FAILED_ROWS_COUNT = "failed_rows_count"


class UserDefinedFailedRowsCheck(Check):
    """
    Eg:

    checks:
      - "Customers must have cst_size":
          failed rows query: |
            SELECT *
            FROM {table_name}
            WHERE cst_size < 0
    """

    def __init__(
        self, check_cfg: "UserDefinedFailedRowsCheckCfg", data_source_scan: "DataSourceScan", partition: "Partition"
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
        )

        from soda.sodacl.user_defined_failed_rows_check_cfg import (
            UserDefinedFailedRowsCheckCfg,
        )

        check_cfg: UserDefinedFailedRowsCheckCfg = self.check_cfg
        self.check_value = None

        metric = UserDefinedFailedRowsMetric(
            data_source_scan=self.data_source_scan,
            check_name=check_cfg.source_line,
            query=check_cfg.query,
            check=self,
            partition=partition,
        )
        metric = self.data_source_scan.resolve_metric(metric)
        self.metrics[KEY_FAILED_ROWS_COUNT] = metric

    def evaluate(self, metrics: Dict[str, Metric], historic_values: Dict[str, object]):
        metric = metrics.get(KEY_FAILED_ROWS_COUNT)
        failed_row_count: int = metric.value
        self.check_value: int = metrics.get(KEY_FAILED_ROWS_COUNT).value

        self.outcome = CheckOutcome.PASS
        if failed_row_count > 0:
            self.outcome = CheckOutcome.FAIL

        self.failed_rows_sample_ref = metric.failed_rows_sample_ref



from __future__ import annotations


class FormatHelper:
    @classmethod
    def is_numeric(cls, format):
        return format in [
            "integer",
            "positive integer",
            "negative integer",
            "decimal",
            "positive decimal",
            "negative decimal",
            "decimal point",
            "positive decimal point",
            "negative decimal point",
            "decimal comma",
            "positive decimal comma",
            "negative decimal comma",
            "percentage",
            "positive percentage",
            "negative percentage",
            "percentage point",
            "positive percentage point",
            "negative percentage point",
            "percentage comma",
            "positive percentage comma",
            "negative percentage comma",
            "money",
            "money point",
            "money comma",
        ]



import logging
import os
import platform
from distutils.util import strtobool
from typing import Dict

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor
from opentelemetry.semconv.resource import ResourceAttributes
from soda.__version__ import SODA_CORE_VERSION
from soda.common.config_helper import ConfigHelper

# from soda.execution.data_source import DataSource
from soda.telemetry.memory_span_exporter import MemorySpanExporter
from soda.telemetry.soda_exporter import SodaConsoleSpanExporter, SodaOTLPSpanExporter

logger = logging.getLogger(__name__)


class SodaTelemetry:
    """Main entry point for Open Telemetry tracing.

    For more info about what and why visit https://github.com/sodadata/soda-sql/issues/543.

    The main goal of this class is to concentrate as much tracing data and logic as reasonable.
    This code design means that compromises have been made on code design in order to facilitate maximum
    transparency and avoid scattering tracing code around the codebase.

    With that being said, some tracing is still present in other files, e.g.:
    - `/soda/core/soda/cli/cli.py` - CLI arguments and options tracing
    - `/soda/core/soda/execution/data_source.py` - safe data source type and hash tracing

    This list is not necessarily exhaustive, search for `from soda.telemetry.soda_telemetry import SodaTelemetry` imports OR
    `set_attribute` method usage to obtain the full list.
    """

    ENDPOINT = "https://collect.soda.io/v1/traces"
    __instance = None
    soda_config = ConfigHelper.get_instance()

    @staticmethod
    def get_instance(test_mode: bool = False):
        if test_mode:
            SodaTelemetry.__instance = None

        if SodaTelemetry.__instance is None:
            SodaTelemetry(test_mode=test_mode)
        return SodaTelemetry.__instance

    def __init__(self, test_mode: bool):
        if SodaTelemetry.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            SodaTelemetry.__instance = self

        self.__send = self.soda_config.send_anonymous_usage_stats or test_mode

        if self.__send:
            logger.info("Setting up usage telemetry.")

            self.__provider = TracerProvider(
                resource=Resource.create(
                    {
                        "os.architecture": platform.architecture(),
                        "python.version": platform.python_version(),
                        "python.implementation": platform.python_implementation(),
                        ResourceAttributes.OS_TYPE: platform.system(),
                        ResourceAttributes.OS_VERSION: platform.version(),
                        "platform": platform.platform(),
                        ResourceAttributes.SERVICE_VERSION: SODA_CORE_VERSION,
                        ResourceAttributes.SERVICE_NAME: "soda",
                        ResourceAttributes.SERVICE_NAMESPACE: "soda-core",
                    }
                )
            )

            if test_mode:
                self.__setup_for_test()
            else:
                self.__setup()
        else:
            logger.info("Skipping usage telemetry setup.")

    def __setup(self):
        """Set up Open Telemetry processors and exporters for normal use."""
        local_debug_mode = self.soda_config.get_value("telemetry_local_debug_mode") or bool(
            strtobool(os.getenv("telemetry_local_debug_mode", "false"))
        )

        if local_debug_mode or logger.getEffectiveLevel() == logging.DEBUG:
            self.__provider.add_span_processor(BatchSpanProcessor(SodaConsoleSpanExporter()))

        if not local_debug_mode:
            otlp_exporter = SodaOTLPSpanExporter(endpoint=self.ENDPOINT)
            otlp_processor = BatchSpanProcessor(otlp_exporter)
            self.__provider.add_span_processor(otlp_processor)

        trace.set_tracer_provider(self.__provider)

    def __setup_for_test(self):
        """Set up Open Telemetry processors and exporters for usage in tests."""
        self.__provider.add_span_processor(SimpleSpanProcessor(MemorySpanExporter.get_instance()))

        trace.set_tracer_provider(self.__provider)

    def get_attribute(self, key: str, default_value=None) -> any:
        """There is no obvious easy way to get attribute value from the current span."""
        raise NotImplementedError()

    def set_attribute(self, key: str, value: str) -> None:
        """Set attribute value in the current span."""
        if self.__send:
            current_span = trace.get_current_span()
            current_span.set_attribute(key, value)

    def set_attributes(self, values: Dict[str, str]) -> None:
        """Set attributes the current span."""
        if self.__send:
            current_span = trace.get_current_span()
            current_span.set_attributes(values)

    @staticmethod
    def obtain_datasource_hash(data_source: "DataSource"):
        return data_source.generate_hash_safe()

    @property
    def user_cookie_id(self) -> str:
        return self.soda_config.get_value("user_cookie_id")



from __future__ import annotations

import re


def string_matches_simple_pattern(input: str, pattern: str) -> bool:
    if "*" in pattern:
        pattern = pattern.replace("*", ".*")
    result = re.fullmatch(pattern, input, re.IGNORECASE)

    return bool(result)



from typing import Dict, List

from soda.execution.check.anomaly_metric_check import AnomalyMetricCheck
from soda.execution.check.check import Check
from soda.execution.check.schema_check import SchemaCheck
from soda.execution.data_source_scan import DataSourceScan
from soda.execution.partition import Partition
from soda.sodacl.anomaly_metric_check_cfg import AnomalyMetricCheckCfg
from soda.sodacl.data_source_check_cfg import DataSourceCheckCfg
from soda.sodacl.schema_check_cfg import SchemaCheckCfg, SchemaValidations
from soda.sodacl.threshold_cfg import ThresholdCfg


class AutomatedMonitoringRun:
    def __init__(self, data_source_scan: DataSourceScan, data_source_check_cfg: DataSourceCheckCfg):
        self.data_source_scan: DataSourceScan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.data_source_check_cfg: DataSourceCheckCfg = data_source_check_cfg
        self.logs = self.data_source_scan.scan._logs
        self.table_names = self._get_table_names()

    def run(self) -> List[Check]:
        automated_checks: List[Check] = []
        annomaly_detection_checks: List[AnomalyMetricCheck] = self.create_anomaly_detection_checks()
        schema_checks: List[SchemaCheck] = self.create_schema_checks()

        automated_checks.extend(annomaly_detection_checks)
        automated_checks.extend(schema_checks)
        return automated_checks

    def create_anomaly_detection_checks(self) -> List[AnomalyMetricCheck]:
        annomaly_detection_checks = []

        for measured_table_name in self.table_names:
            anomaly_metric_check_cfg = AnomalyMetricCheckCfg(
                source_header=f"checks for {measured_table_name}",
                source_line="anomaly score for row_count < default",
                source_configurations=None,
                location=self.data_source_check_cfg.location,
                name=None,
                metric_name="row_count",
                metric_args=None,
                missing_and_valid_cfg=None,
                filter=None,
                condition=None,
                metric_expression=None,
                metric_query=None,
                change_over_time_cfg=None,
                fail_threshold_cfg=None,
                warn_threshold_cfg=ThresholdCfg(gt=0.9),
                is_automated_monitoring=True,
            )

            # Mock partition
            table = self.data_source_scan.get_or_create_table(measured_table_name)
            partition: Partition = table.get_or_create_partition(None)
            anomaly_metric_check = AnomalyMetricCheck(
                anomaly_metric_check_cfg, self.data_source_scan, partition=partition
            )
            anomaly_metric_check.archetype = "volumeConsistency"

            # Execute query to change the value of metric class to get the historical results
            self.data_source_scan.execute_queries()

            annomaly_detection_checks.append(anomaly_metric_check)

        return annomaly_detection_checks

    def create_schema_checks(self) -> List[SchemaCheck]:
        schema_checks = []

        fail_validations = SchemaValidations(
            required_column_names=None,
            required_column_types=None,
            required_column_indexes=None,
            forbidden_column_names=None,
            is_column_addition_forbidden=False,
            is_column_deletion_forbidden=True,
            is_column_type_change_forbidden=True,
            is_column_index_change_forbidden=True,
        )

        warn_validations = SchemaValidations(
            required_column_names=None,
            required_column_types=None,
            required_column_indexes=None,
            forbidden_column_names=None,
            is_column_addition_forbidden=True,
            is_column_deletion_forbidden=False,
            is_column_type_change_forbidden=False,
            is_column_index_change_forbidden=False,
        )

        for measured_table_name in self.table_names:
            schema_check_cfg = SchemaCheckCfg(
                source_header=f"checks for {measured_table_name}",
                source_line="schema",
                source_configurations=None,
                location=self.data_source_check_cfg.location,
                name=None,
                warn_validations=warn_validations,
                fail_validations=fail_validations,
                is_automated_monitoring=True,
            )

            # Mock partition
            table = self.data_source_scan.get_or_create_table(measured_table_name)
            partition: Partition = table.get_or_create_partition(None)
            schema_check = SchemaCheck(schema_check_cfg, self.data_source_scan, partition=partition)
            schema_check.archetype = "schemaConsistency"

            # Execute query to change the value of metric class to get the historical results
            self.data_source_scan.execute_queries()
            schema_checks.append(schema_check)
        return schema_checks

    def _get_table_names(self) -> Dict[str, Dict[str, str]]:
        """
        Returns a dict that maps table names to a dict that maps column names to column types.
        {table_name -> {column_name -> column_type}}
        """
        include_tables = self.data_source_check_cfg.include_tables
        exclude_tables = self.data_source_check_cfg.exclude_tables
        table_names = self.data_source.get_table_names(include_tables=include_tables, exclude_tables=exclude_tables)
        return table_names



from soda.execution.metric.query_metric import QueryMetric


class SchemaMetric(QueryMetric):
    def __init__(
        self,
        data_source_scan: "DataSourceScan",
        partition: "Partition",
        check: "Check",
    ):
        super().__init__(
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
            name="schema",
            check=check,
            identity_parts=[],
        )

    def ensure_query(self):
        self.schema_query = self.data_source_scan.data_source.create_table_columns_query(self.partition, self)
        self.queries.append(self.schema_query)
        self.data_source_scan.queries.append(self.schema_query)

    def get_cloud_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        return {
            "identity": self.identity,
            "metricName": self.name,
            "dataSourceName": self.data_source_scan.data_source.data_source_name,
            "tableName": Partition.get_table_name(self.partition),
            "partitionName": Partition.get_partition_name(self.partition),
            "columnName": Column.get_partition_name(self.column),
            "value": [{"columnName": c["name"], "sourceDataType": c["type"]} for c in self.value],
        }

    def get_dict(self):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        return {
            "identity": self.identity,
            "metricName": self.name,
            "dataSourceName": self.data_source_scan.data_source.data_source_name,
            "tableName": Partition.get_table_name(self.partition),
            "partitionName": Partition.get_partition_name(self.partition),
            "columnName": Column.get_partition_name(self.column),
            "value": [{"columnName": c["name"], "sourceDataType": c["type"]} for c in self.value],
        }



from soda.execution.table import Table


class Column:
    def __init__(self, table: "Table", column_name: str):
        from soda.sodacl.column_configurations_cfg import ColumnConfigurationsCfg

        self.data_source_scan = table.data_source_scan
        self.table = table
        self.column_name = str(column_name)
        self.column_configurations_cfg: ColumnConfigurationsCfg = None

    def set_column_configuration_cfg(self, column_configurations_cfg: "ColumnConfigurationsCfg"):
        self.column_configurations_cfg = column_configurations_cfg

    @classmethod
    def get_partition_name(cls, column):
        return column.column_name if isinstance(column, Column) else None



import ast
import inspect
import textwrap
from functools import wraps
from typing import Dict, Optional

from opentelemetry import trace
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from opentelemetry.trace.span import Span
from opentelemetry.trace.status import Status, StatusCode
from soda.telemetry.soda_telemetry import SodaTelemetry

trace_context_propagator = TraceContextTextMapPropagator()
trace_context_carrier = {}

soda_telemetry = SodaTelemetry.get_instance()
tracer = trace.get_tracer_provider().get_tracer(__name__)


def get_decorators(function):
    """
    Fancy introspection - very WIP
    """
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = {}
        for n in node.decorator_list:
            print(ast.dump(n))
            name = ""
            if isinstance(n, ast.Call):
                group = n.func.value.id
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id

                for a in n.args:
                    print(ast.dump(a))
            else:
                group = n.attr if isinstance(n, ast.Attribute) else n.id
                name = None

            if group not in decorators[node.name]:
                decorators[node.name][group] = []

            if name:
                decorators[node.name][group].append({name})

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(textwrap.dedent(inspect.getsource(function))))
    return decorators


def soda_trace(fn: callable):
    def _before_exec(span: Span, fn: callable):
        span.set_attribute("user_cookie_id", soda_telemetry.user_cookie_id)

    def _after_exec(span: Span, error: Optional[BaseException] = None):
        span.set_status(Status(StatusCode.OK))
        if str(error) == "3":
            # Only error code 3 means actual execution error, 1 and 2 are reserved for other use.
            span.set_status(Status(StatusCode.ERROR))

    @wraps(fn)
    def wrapper(*original_args, **original_kwargs):
        ctx = trace_context_propagator.extract(carrier=trace_context_carrier)
        with tracer.start_as_current_span(f"{fn.__module__}.{fn.__name__}", context=ctx) as span:
            trace_context_propagator.inject(carrier=trace_context_carrier)
            result = None
            try:
                _before_exec(span, fn)
                result = fn(*original_args, **original_kwargs)
                span.set_status(Status(StatusCode.OK))
                _after_exec(span)
            except BaseException as e:
                # Catch base exception to deal with system exit codes as well.
                _after_exec(span, e)
                # A bit of a workaround - re-raise the error so that stacktrace is intact. Open telemetry takes care of setting the error code.
                raise e
        return result

    return wrapper


def span_setup_function_args(args: Dict):
    for prefix, values in args.items():
        for key, value in values.items():
            soda_telemetry.set_attribute(f"{prefix}_{key}", value or "")



# Generated from /Users/vijay/work/soda/code/soda-core/soda/core/soda/sodacl/antlr/SodaCLAntlr.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .SodaCLAntlrParser import SodaCLAntlrParser
else:
    from SodaCLAntlrParser import SodaCLAntlrParser

# This class defines a complete listener for a parse tree produced by SodaCLAntlrParser.
class SodaCLAntlrListener(ParseTreeListener):

    # Enter a parse tree produced by SodaCLAntlrParser#check.
    def enterCheck(self, ctx:SodaCLAntlrParser.CheckContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#check.
    def exitCheck(self, ctx:SodaCLAntlrParser.CheckContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#freshness_check.
    def enterFreshness_check(self, ctx:SodaCLAntlrParser.Freshness_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#freshness_check.
    def exitFreshness_check(self, ctx:SodaCLAntlrParser.Freshness_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#freshness_variable.
    def enterFreshness_variable(self, ctx:SodaCLAntlrParser.Freshness_variableContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#freshness_variable.
    def exitFreshness_variable(self, ctx:SodaCLAntlrParser.Freshness_variableContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#warn_qualifier.
    def enterWarn_qualifier(self, ctx:SodaCLAntlrParser.Warn_qualifierContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#warn_qualifier.
    def exitWarn_qualifier(self, ctx:SodaCLAntlrParser.Warn_qualifierContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#failed_rows_check.
    def enterFailed_rows_check(self, ctx:SodaCLAntlrParser.Failed_rows_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#failed_rows_check.
    def exitFailed_rows_check(self, ctx:SodaCLAntlrParser.Failed_rows_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#group_by_check.
    def enterGroup_by_check(self, ctx:SodaCLAntlrParser.Group_by_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#group_by_check.
    def exitGroup_by_check(self, ctx:SodaCLAntlrParser.Group_by_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#row_count_comparison_check.
    def enterRow_count_comparison_check(self, ctx:SodaCLAntlrParser.Row_count_comparison_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#row_count_comparison_check.
    def exitRow_count_comparison_check(self, ctx:SodaCLAntlrParser.Row_count_comparison_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#metric_check.
    def enterMetric_check(self, ctx:SodaCLAntlrParser.Metric_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#metric_check.
    def exitMetric_check(self, ctx:SodaCLAntlrParser.Metric_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#default_anomaly_threshold.
    def enterDefault_anomaly_threshold(self, ctx:SodaCLAntlrParser.Default_anomaly_thresholdContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#default_anomaly_threshold.
    def exitDefault_anomaly_threshold(self, ctx:SodaCLAntlrParser.Default_anomaly_thresholdContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#change_over_time.
    def enterChange_over_time(self, ctx:SodaCLAntlrParser.Change_over_timeContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#change_over_time.
    def exitChange_over_time(self, ctx:SodaCLAntlrParser.Change_over_timeContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#change_over_time_config.
    def enterChange_over_time_config(self, ctx:SodaCLAntlrParser.Change_over_time_configContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#change_over_time_config.
    def exitChange_over_time_config(self, ctx:SodaCLAntlrParser.Change_over_time_configContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#change_aggregation.
    def enterChange_aggregation(self, ctx:SodaCLAntlrParser.Change_aggregationContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#change_aggregation.
    def exitChange_aggregation(self, ctx:SodaCLAntlrParser.Change_aggregationContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#same_day_last_week.
    def enterSame_day_last_week(self, ctx:SodaCLAntlrParser.Same_day_last_weekContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#same_day_last_week.
    def exitSame_day_last_week(self, ctx:SodaCLAntlrParser.Same_day_last_weekContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#percent.
    def enterPercent(self, ctx:SodaCLAntlrParser.PercentContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#percent.
    def exitPercent(self, ctx:SodaCLAntlrParser.PercentContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#anomaly_score.
    def enterAnomaly_score(self, ctx:SodaCLAntlrParser.Anomaly_scoreContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#anomaly_score.
    def exitAnomaly_score(self, ctx:SodaCLAntlrParser.Anomaly_scoreContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#metric.
    def enterMetric(self, ctx:SodaCLAntlrParser.MetricContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#metric.
    def exitMetric(self, ctx:SodaCLAntlrParser.MetricContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#metric_name.
    def enterMetric_name(self, ctx:SodaCLAntlrParser.Metric_nameContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#metric_name.
    def exitMetric_name(self, ctx:SodaCLAntlrParser.Metric_nameContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#metric_args.
    def enterMetric_args(self, ctx:SodaCLAntlrParser.Metric_argsContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#metric_args.
    def exitMetric_args(self, ctx:SodaCLAntlrParser.Metric_argsContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#metric_arg.
    def enterMetric_arg(self, ctx:SodaCLAntlrParser.Metric_argContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#metric_arg.
    def exitMetric_arg(self, ctx:SodaCLAntlrParser.Metric_argContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#threshold.
    def enterThreshold(self, ctx:SodaCLAntlrParser.ThresholdContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#threshold.
    def exitThreshold(self, ctx:SodaCLAntlrParser.ThresholdContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#between_threshold.
    def enterBetween_threshold(self, ctx:SodaCLAntlrParser.Between_thresholdContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#between_threshold.
    def exitBetween_threshold(self, ctx:SodaCLAntlrParser.Between_thresholdContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#comparator_threshold.
    def enterComparator_threshold(self, ctx:SodaCLAntlrParser.Comparator_thresholdContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#comparator_threshold.
    def exitComparator_threshold(self, ctx:SodaCLAntlrParser.Comparator_thresholdContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#zones_threshold.
    def enterZones_threshold(self, ctx:SodaCLAntlrParser.Zones_thresholdContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#zones_threshold.
    def exitZones_threshold(self, ctx:SodaCLAntlrParser.Zones_thresholdContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#outcome.
    def enterOutcome(self, ctx:SodaCLAntlrParser.OutcomeContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#outcome.
    def exitOutcome(self, ctx:SodaCLAntlrParser.OutcomeContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#zone_comparator.
    def enterZone_comparator(self, ctx:SodaCLAntlrParser.Zone_comparatorContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#zone_comparator.
    def exitZone_comparator(self, ctx:SodaCLAntlrParser.Zone_comparatorContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#comparator.
    def enterComparator(self, ctx:SodaCLAntlrParser.ComparatorContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#comparator.
    def exitComparator(self, ctx:SodaCLAntlrParser.ComparatorContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#threshold_value.
    def enterThreshold_value(self, ctx:SodaCLAntlrParser.Threshold_valueContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#threshold_value.
    def exitThreshold_value(self, ctx:SodaCLAntlrParser.Threshold_valueContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#freshness_threshold_value.
    def enterFreshness_threshold_value(self, ctx:SodaCLAntlrParser.Freshness_threshold_valueContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#freshness_threshold_value.
    def exitFreshness_threshold_value(self, ctx:SodaCLAntlrParser.Freshness_threshold_valueContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#reference_check.
    def enterReference_check(self, ctx:SodaCLAntlrParser.Reference_checkContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#reference_check.
    def exitReference_check(self, ctx:SodaCLAntlrParser.Reference_checkContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#source_column_name.
    def enterSource_column_name(self, ctx:SodaCLAntlrParser.Source_column_nameContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#source_column_name.
    def exitSource_column_name(self, ctx:SodaCLAntlrParser.Source_column_nameContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#target_column_name.
    def enterTarget_column_name(self, ctx:SodaCLAntlrParser.Target_column_nameContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#target_column_name.
    def exitTarget_column_name(self, ctx:SodaCLAntlrParser.Target_column_nameContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#section_header.
    def enterSection_header(self, ctx:SodaCLAntlrParser.Section_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#section_header.
    def exitSection_header(self, ctx:SodaCLAntlrParser.Section_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#table_checks_header.
    def enterTable_checks_header(self, ctx:SodaCLAntlrParser.Table_checks_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#table_checks_header.
    def exitTable_checks_header(self, ctx:SodaCLAntlrParser.Table_checks_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#partition_name.
    def enterPartition_name(self, ctx:SodaCLAntlrParser.Partition_nameContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#partition_name.
    def exitPartition_name(self, ctx:SodaCLAntlrParser.Partition_nameContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#table_filter_header.
    def enterTable_filter_header(self, ctx:SodaCLAntlrParser.Table_filter_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#table_filter_header.
    def exitTable_filter_header(self, ctx:SodaCLAntlrParser.Table_filter_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#column_configurations_header.
    def enterColumn_configurations_header(self, ctx:SodaCLAntlrParser.Column_configurations_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#column_configurations_header.
    def exitColumn_configurations_header(self, ctx:SodaCLAntlrParser.Column_configurations_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#checks_for_each_dataset_header.
    def enterChecks_for_each_dataset_header(self, ctx:SodaCLAntlrParser.Checks_for_each_dataset_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#checks_for_each_dataset_header.
    def exitChecks_for_each_dataset_header(self, ctx:SodaCLAntlrParser.Checks_for_each_dataset_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#checks_for_each_column_header.
    def enterChecks_for_each_column_header(self, ctx:SodaCLAntlrParser.Checks_for_each_column_headerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#checks_for_each_column_header.
    def exitChecks_for_each_column_header(self, ctx:SodaCLAntlrParser.Checks_for_each_column_headerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#signed_number.
    def enterSigned_number(self, ctx:SodaCLAntlrParser.Signed_numberContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#signed_number.
    def exitSigned_number(self, ctx:SodaCLAntlrParser.Signed_numberContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#number.
    def enterNumber(self, ctx:SodaCLAntlrParser.NumberContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#number.
    def exitNumber(self, ctx:SodaCLAntlrParser.NumberContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#integer.
    def enterInteger(self, ctx:SodaCLAntlrParser.IntegerContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#integer.
    def exitInteger(self, ctx:SodaCLAntlrParser.IntegerContext):
        pass


    # Enter a parse tree produced by SodaCLAntlrParser#identifier.
    def enterIdentifier(self, ctx:SodaCLAntlrParser.IdentifierContext):
        pass

    # Exit a parse tree produced by SodaCLAntlrParser#identifier.
    def exitIdentifier(self, ctx:SodaCLAntlrParser.IdentifierContext):
        pass



del SodaCLAntlrParser


from __future__ import annotations

from datetime import datetime, timedelta

from soda.common.exception_helper import get_exception_stacktrace
from soda.common.query_helper import parse_columns_from_query
from soda.common.undefined_instance import undefined
from soda.sampler.db_sample import DbSample
from soda.sampler.sample_context import SampleContext
from soda.sampler.sampler import Sampler


class Query:
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        table: Table = None,
        partition: Partition = None,
        column: Column = None,
        unqualified_query_name: str = None,
        sql: str | None = None,
        sample_name: str = "failed_rows",
        location: Location | None = None,
        samples_limit: int | None = None,
    ):
        self.logs = data_source_scan.scan._logs
        self.data_source_scan = data_source_scan
        self.query_name: str = Query.build_query_name(
            data_source_scan, table, partition, column, unqualified_query_name
        )
        self.sample_name = sample_name
        self.table: Table | None = table
        self.partition: Partition | None = partition
        self.column: Column | None = column
        self.location: Location | None = location
        self.samples_limit: int | None = samples_limit

        # The SQL query that is used _fetchone or _fetchall or _store
        # This field can also be initialized in the execute method before any of _fetchone,
        # _fetchall or _store are called
        self.sql: str = data_source_scan.scan.jinja_resolve(sql)

        self.passing_sql: str | None = None
        self.failing_sql: str | None = None

        self.failing_rows_sql_aggregated: str | None = None

        # Following fields are initialized in execute method
        self.description: tuple | None = None
        self.row: tuple | None = None
        self.rows: list[tuple] | None = None
        self.sample_ref: SampleRef | None = None
        self.exception: BaseException | None = None
        self.duration: timedelta | None = None

    def get_cloud_dicts(self) -> list(dict(str, any)):
        dicts = [self.get_dict()]

        if self.failing_sql:
            dicts.append(self.get_dict("failing_sql", self.failing_sql))

        if self.passing_sql:
            dicts.append(self.get_dict("passing_sql", self.passing_sql))

        if self.failing_rows_sql_aggregated:
            dicts.append(self.get_dict("failing_rows_sql_aggregated", self.failing_rows_sql_aggregated))

        return dicts

    def get_dict(self, name_suffix: str | None = None, sql: str | None = None) -> dict:
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        name = self.query_name

        if name_suffix:
            name += f".{name_suffix}"
        return {
            "name": name,
            "dataSource": self.data_source_scan.data_source.data_source_name,
            "table": Partition.get_table_name(self.partition),
            "partition": Partition.get_partition_name(self.partition),
            "column": Column.get_partition_name(self.column),
            "sql": sql or self.sql,
            "exception": None if name_suffix or sql else get_exception_stacktrace(self.exception),
            "duration": None if name_suffix or sql else self.duration,
        }

    @staticmethod
    def build_query_name(data_source_scan, table, partition, column, unqualified_query_name):
        full_query_pieces = [data_source_scan.data_source.data_source_name]
        if partition is not None and partition.partition_name is not None:
            full_query_pieces.append(f"{partition.table.table_name}[{partition.partition_name}]")
        elif table is not None:
            full_query_pieces.append(f"{table.table_name}")
        if column is not None:
            full_query_pieces.append(f"{column.column_name}")
        full_query_pieces.append(unqualified_query_name)
        return ".".join(full_query_pieces)

    def execute(self):
        """
        Execute method implementations should
          - invoke either self.fetchone, self.fetchall or self.store
          - update the metrics with value and optionally other diagnostic information
        """
        # TODO: some of the subclasses couple setting metric with storing the sample - refactor that.
        self.fetchall()

    def fetchone(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        start = datetime.now()
        data_source = self.data_source_scan.data_source
        try:
            cursor = data_source.connection.cursor()
            try:
                self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                cursor.execute(self.sql)
                self.row = cursor.fetchone()
                self.description = cursor.description
            finally:
                cursor.close()
        except BaseException as e:
            self.exception = e
            self.logs.error(
                message=f"Query execution error in {self.query_name}: {e}\n{self.sql}",
                exception=e,
                location=self.location,
            )
            data_source.query_failed(e)
        finally:
            self.duration = datetime.now() - start

    def fetchall(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        start = datetime.now()
        data_source = self.data_source_scan.data_source
        try:
            cursor = data_source.connection.cursor()
            try:
                self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                cursor.execute(self.sql)
                self.rows = cursor.fetchall()
                self.description = cursor.description
            finally:
                cursor.close()
        except BaseException as e:
            self.exception = e
            self.logs.error(f"Query error: {self.query_name}: {e}\n{self.sql}", exception=e, location=self.location)
            data_source.query_failed(e)
        finally:
            self.duration = datetime.now() - start

    def store(self):
        """
        DataSource query execution exceptions will be caught and result in the
        self.exception being populated.
        """
        self.__append_to_scan()
        sampler: Sampler = self.data_source_scan.scan._configuration.sampler
        start = datetime.now()
        data_source = self.data_source_scan.data_source
        try:
            cursor = data_source.connection.cursor()
            try:
                # Check if query does not contain forbidden columns and only create sample if it does not.
                # Query still needs to execute in case this is a query that also sets a metric value. (e.g. reference check)
                allow_samples = True
                offending_columns = []

                if self.partition and self.partition.table:
                    query_columns = parse_columns_from_query(self.sql)

                    for column in query_columns:
                        if self.data_source_scan.data_source.is_column_excluded(
                            self.partition.table.table_name, column
                        ):
                            allow_samples = False
                            offending_columns.append(column)

                # A bit of a hacky workaround for queries that also set the metric in one go.
                # TODO: revisit after decoupling getting metric values and storing samples. This can be dangerous, it sets the metric value
                # only when metric value is not set, but this could cause weird regressions.
                set_metric = False
                if hasattr(self, "metric") and self.metric and self.metric.value == undefined:
                    set_metric = True

                if set_metric or allow_samples:
                    self.logs.debug(f"Query {self.query_name}:\n{self.sql}")
                    cursor.execute(str(self.sql))
                    self.description = cursor.description
                    db_sample = DbSample(cursor, self.data_source_scan.data_source)

                if set_metric:
                    self.metric.set_value(len(db_sample.get_rows()))

                if allow_samples:
                    # TODO Hacky way to get the check name, check name isn't there when dataset samples are taken
                    check_name = next(iter(self.metric.checks)).name if hasattr(self, "metric") else None
                    sample_context = SampleContext(
                        sample=db_sample,
                        sample_name=self.sample_name,
                        query=self.sql,
                        data_source=self.data_source_scan.data_source,
                        partition=self.partition,
                        column=self.column,
                        scan=self.data_source_scan.scan,
                        logs=self.data_source_scan.scan._logs,
                        samples_limit=self.samples_limit,
                        passing_sql=self.passing_sql,
                        check_name=check_name,
                    )

                    self.sample_ref = sampler.store_sample(sample_context)
                else:
                    self.logs.info(
                        f"Skipping samples from query '{self.query_name}'. Excluded column(s) present: {offending_columns}."
                    )
            finally:
                cursor.close()
        except BaseException as e:
            self.exception = e
            self.logs.error(f"Query error: {self.query_name}: {e}\n{self.sql}", exception=e, location=self.location)
            data_source.query_failed(e)
        finally:
            self.duration = datetime.now() - start

    def __append_to_scan(self):
        scan = self.data_source_scan.scan
        self.index = len(scan._queries)
        scan._queries.append(self)



from abc import ABC, abstractmethod
from typing import Tuple

from soda.sampler.sample_schema import SampleSchema


class Sample(ABC):
    @abstractmethod
    def get_rows(self) -> Tuple[Tuple]:
        pass

    @abstractmethod
    def get_schema(self) -> SampleSchema:
        pass



from __future__ import annotations

import copy

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.check_type import CheckType
from soda.execution.metric.metric import Metric
from soda.execution.partition import Partition

GROUP_BY_RESULTS = "group_by_results"


class GroupByCheck(Check):
    def __init__(
        self,
        check_cfg: GroupByCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
        )

        self.check_value = None
        self.check_type = CheckType.LOCAL

        from soda.execution.metric.group_by_metric import GroupByMetric

        group_by_metric = data_source_scan.resolve_metric(
            GroupByMetric(
                data_source_scan=self.data_source_scan,
                partition=partition,
                query=self.check_cfg.query,
                check=self,
            )
        )
        self.metrics[GROUP_BY_RESULTS] = group_by_metric

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        query_results = metrics[GROUP_BY_RESULTS].value
        group_limit = self.check_cfg.group_limit

        if len(query_results) > group_limit:
            raise Exception(
                f"Total number of groups {len(query_results)} exceeds configured group limit: {group_limit}"
            )

        fields = list(self.check_cfg.fields)
        group_check_cfgs = self.check_cfg.check_cfgs
        groups = [tuple(map(qr.get, fields)) for qr in query_results]

        group_checks = []

        for group in groups:
            for gcc in group_check_cfgs:
                if gcc.name is None:
                    raise Exception("name property is required for the group check")

                group_name = f"{','.join(str(v) for v in group)}"
                config = copy.deepcopy(gcc)
                config.name = gcc.name + f" [{group_name}]"
                config.source_configurations["group_value"] = f"[{group_name}]"
                column = ",".join(fields)
                gc = Check.create(
                    check_cfg=config, data_source_scan=self.data_source_scan, partition=self.partition, column=column
                )
                result = next(filter(lambda qr: tuple(map(qr.get, fields)) == group, query_results))
                if result is not None:
                    gc.check_value = result[config.metric_name]
                    metric = Metric(
                        self.data_source_scan,
                        self.partition,
                        column=None,
                        name=config.name,
                        check=None,
                        identity_parts=[],
                    )

                    # TODO fetch historic values, change over time checks will not work yet
                    # historic_values = {}
                    # if gc.historic_descriptors:
                    #     for hd_key, hd in gc.historic_descriptors.items():
                    #         print(f"hd_key: {hd_key}, hd: {hd}")
                    #         historic_values[hd_key] = self.data_source_scan.scan.__get_historic_data_from_soda_cloud_metric_store(hd)

                    metric.set_value(gc.check_value)
                    self.data_source_scan.scan._add_metric(metric)
                    gc.metrics = {config.metric_name: metric}
                    gc.evaluate(metrics=None, historic_values=None)

                    cloud_group_attr = {
                        "group": {
                            "identity": self.create_identity(with_datasource=True, with_filename=True),
                            "name": gcc.name,
                            "distinctLabel": group_name,
                        }
                    }
                    gc.cloud_dict.update(cloud_group_attr)
                group_checks.append(gc)

        self.data_source_scan.scan._checks.extend(group_checks)

        if all(gc.outcome == CheckOutcome.PASS for gc in group_checks):
            self.outcome = CheckOutcome.PASS
        elif any(gc.outcome == CheckOutcome.FAIL for gc in group_checks):
            self.outcome = CheckOutcome.FAIL
        else:
            self.outcome = CheckOutcome.PASS

    def get_cloud_diagnostics_dict(self) -> dict:
        group_by_diagnostics = {}
        return group_by_diagnostics

    def get_log_diagnostic_lines(self) -> list[str]:
        diagnostics_texts: list[str] = []
        return diagnostics_texts



from typing import List

from soda.profiling.discover_table_result_table import DiscoverTablesResultTable
from soda.sodacl.data_source_check_cfg import DataSourceCheckCfg


class DiscoverTablesResult:
    def __init__(self, data_source_check_cfg: DataSourceCheckCfg):
        self.data_source_check_cfg: DataSourceCheckCfg = data_source_check_cfg
        self.tables: List[DiscoverTablesResultTable] = []

    def create_table(self, table_name: str, data_source_name: str) -> DiscoverTablesResultTable:
        table = DiscoverTablesResultTable(table_name, data_source_name)
        self.tables.append(table)
        return table



from __future__ import annotations

from copy import deepcopy

from soda.sodacl.location import Location


class CheckCfg:
    """Represents the config from the check config (yaml or otherwise) as closely as possible."""

    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: dict | None,
        location: Location,
        name: str | None,
        samples_limit: int | None = None,
    ):
        self.source_header: str = source_header
        self.source_line: str = source_line
        self.source_configurations = source_configurations
        self.location: Location = location
        self.name: str | None = name
        self.samples_limit: int | None = samples_limit

    def get_column_name(self) -> str | None:
        pass

    def instantiate_for_each_dataset(
        self, name: str, table_alias: str, table_name: str, partition_name: str
    ) -> CheckCfg:
        instantiated_check_cfg = deepcopy(self)
        partition_replace = f" [{partition_name}]" if partition_name else ""
        instantiated_check_cfg.name = name
        instantiated_check_cfg.source_header = f"checks for {table_alias} being {table_name}{partition_replace}"
        return instantiated_check_cfg



from __future__ import annotations

from typing import TYPE_CHECKING

from soda.execution.query.query import Query
from soda.profiling.profile_columns_result import ProfileColumnsResultColumn

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan
    from soda.sodacl.data_source_check_cfg import ProfileColumnsCfg


class NumericColumnProfiler:
    def __init__(
        self,
        data_source_scan: DataSourceScan,
        profile_columns_cfg: ProfileColumnsCfg,
        table_name: str,
        column_name: str,
        column_data_type: str,
    ) -> None:
        self.data_source_scan = data_source_scan
        self.data_source = data_source_scan.data_source
        self.logs = data_source_scan.scan._logs
        self.profile_columns_cfg = profile_columns_cfg
        self.table_name = table_name
        self.column_name = column_name
        self.column_data_type = column_data_type
        self.result_column = ProfileColumnsResultColumn(column_name=column_name, column_data_type=column_data_type)

    def profile(self) -> ProfileColumnsResultColumn:
        self.logs.debug(f"Profiling column {self.column_name} of {self.table_name}")

        # mins, maxs, min, max, frequent values
        self._set_result_column_value_frequency_attributes()

        # Average, sum, variance, standard deviation, distinct values, missing values
        self._set_result_column_aggregation_attributes()

        # histogram
        self._set_result_column_histogram_attributes()
        return self.result_column

    def _set_result_column_value_frequency_attributes(self) -> None:
        value_frequencies = self._compute_value_frequency()
        if value_frequencies:
            self.result_column.set_min_max_metrics(value_frequencies=value_frequencies)
            self.result_column.set_frequency_metric(value_frequencies=value_frequencies)
        else:
            self.logs.error(
                "Database returned no results for minumum values, maximum values and "
                f"frequent values in table: {self.table_name}, columns: {self.column_name}"
            )

    def _set_result_column_aggregation_attributes(self) -> None:
        aggregated_metrics = self._compute_aggregated_metrics()
        if aggregated_metrics:
            self.result_column.set_numeric_aggregation_metrics(aggregated_metrics=aggregated_metrics)
        else:
            self.logs.error(
                f"Database returned no results for aggregates in table: {self.table_name}, columns: {self.column_name}"
            )

    def _set_result_column_histogram_attributes(self) -> None:
        histogram_values = self._compute_histogram()
        if histogram_values:
            self.result_column.set_histogram(histogram_values=histogram_values)
        else:
            self.logs.error(
                f"Database returned no results for histograms in table: {self.table_name}, columns: {self.column_name}"
            )

    def _compute_value_frequency(self) -> list[tuple] | None:
        value_frequencies_sql = self.data_source.profiling_sql_values_frequencies_query(
            "numeric",
            self.table_name,
            self.column_name,
            self.profile_columns_cfg.limit_mins_maxs,
            self.profile_columns_cfg.limit_frequent_values,
        )

        value_frequencies_query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"profiling-{self.table_name}-{self.column_name}-value-frequencies-numeric",
            sql=value_frequencies_sql,
        )
        value_frequencies_query.execute()
        rows = value_frequencies_query.rows
        return rows

    def _compute_aggregated_metrics(self) -> list[tuple] | None:
        aggregates_sql = self.data_source.profiling_sql_aggregates_numeric(self.table_name, self.column_name)
        aggregates_query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"profiling-{self.table_name}-{self.column_name}-profiling-aggregates",
            sql=aggregates_sql,
        )
        aggregates_query.execute()
        rows = aggregates_query.rows
        return rows

    def _compute_histogram(self) -> None | dict[str, list]:
        if self.result_column.min is None:
            self.logs.warning("Min cannot be None, make sure the min metric is derived before histograms")
        if self.result_column.max is None:
            self.logs.warning("Max cannot be None, make sure the min metric is derived before histograms")
        if self.result_column.distinct_values is None:
            self.logs.warning(
                "Distinct values cannot be None, make sure the distinct values metric is derived before histograms"
            )
        if (
            self.result_column.min is None
            or self.result_column.max is None
            or self.result_column.distinct_values is None
        ):
            self.logs.warning(
                f"Histogram query for {self.table_name}, column {self.column_name} skipped. See earlier warnings."
            )
            return None

        histogram_sql, bins_list = self.data_source.histogram_sql_and_boundaries(
            table_name=self.table_name,
            column_name=self.column_name,
            min_value=self.result_column.min,
            max_value=self.result_column.max,
            n_distinct=self.result_column.distinct_values,
            column_type=self.column_data_type,
        )
        if histogram_sql is None:
            return None

        histogram_query = Query(
            data_source_scan=self.data_source_scan,
            unqualified_query_name=f"profiling-{self.table_name}-{self.column_name}-histogram",
            sql=histogram_sql,
        )
        histogram_query.execute()
        histogram_values = histogram_query.rows

        if histogram_values is None:
            return None
        histogram = {}
        histogram["boundaries"] = bins_list
        histogram["frequencies"] = [int(freq) if freq is not None else 0 for freq in histogram_values[0]]
        return histogram



from __future__ import annotations

import logging

from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef
from soda.sampler.sampler import Sampler

logger = logging.getLogger(__name__)


class DefaultSampler(Sampler):
    def store_sample(self, sample_context: SampleContext) -> SampleRef:
        self.logs.info("Using DefaultSampler")
        sample_rows = sample_context.sample.get_rows()
        row_count = len(sample_rows)

        sample_schema = sample_context.sample.get_schema()

        return SampleRef(
            name=sample_context.sample_name,
            schema=sample_schema,
            total_row_count=row_count,
            stored_row_count=row_count,
            type=SampleRef.TYPE_NOT_PERSISTED,
            message="Samples are not sent to Soda Cloud",
        )



from __future__ import annotations

from soda.sampler.sample_context import SampleContext
from soda.sampler.sample_ref import SampleRef
from soda.sampler.sampler import Sampler


class SodaCloudSampler(Sampler):
    def store_sample(self, sample_context: SampleContext) -> SampleRef | None:
        self.logs.info(f"Sending failed row samples to Soda Cloud")
        sample_rows = sample_context.sample.get_rows()
        row_count = len(sample_rows)
        sample_schema = sample_context.sample.get_schema()

        if row_count == 0:
            type = SampleRef.TYPE_NOT_PERSISTED
            soda_cloud_file_id = None
        else:
            type = "soda_cloud"
            scan = sample_context.scan
            soda_cloud = scan._configuration.soda_cloud
            soda_cloud_file_id = soda_cloud.upload_sample(
                scan=scan,
                sample_rows=sample_rows,
                sample_file_name=sample_context.get_sample_file_name(),
                samples_limit=sample_context.samples_limit,
            )
        if sample_context.samples_limit is not None:
            stored_row_count = row_count if row_count < sample_context.samples_limit else sample_context.samples_limit
        else:
            stored_row_count = row_count

        return SampleRef(
            name=sample_context.sample_name,
            schema=sample_schema,
            total_row_count=row_count,
            stored_row_count=stored_row_count,
            type=type,
            soda_cloud_file_id=soda_cloud_file_id,
        )



from __future__ import annotations

import re
from dataclasses import dataclass, field

from soda.execution.check.check import Check
from soda.execution.check_outcome import CheckOutcome
from soda.execution.metric.metric import Metric
from soda.execution.partition import Partition
from soda.execution.schema_comparator import SchemaComparator
from soda.soda_cloud.historic_descriptor import HistoricChangeOverTimeDescriptor
from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg
from soda.sodacl.schema_check_cfg import SchemaCheckCfg, SchemaValidations

KEY_SCHEMA_MEASURED = "schema_measured"
KEY_SCHEMA_PREVIOUS = "schema_previous"


@dataclass
class SchemaCheckValidationResult:
    missing_column_names: list[str] = field(default_factory=list)
    present_column_names: list[str] = field(default_factory=list)
    column_type_mismatches: dict[str, str] = field(default_factory=dict)
    column_index_mismatches: dict[str, str] = field(default_factory=dict)
    column_additions: list = field(default_factory=list)
    column_deletions: list = field(default_factory=list)
    column_type_changes: dict = field(default_factory=dict)
    column_index_changes: dict = field(default_factory=dict)


class SchemaCheck(Check):
    def __init__(
        self,
        check_cfg: CheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=None,
        )

        self.measured_schema: list[dict[str, str]] | None = None
        self.schema_comparator = None

        self.warn_result: SchemaCheckValidationResult | None = None
        self.fail_result: SchemaCheckValidationResult | None = None

        from soda.execution.metric.schema_metric import SchemaMetric

        schema_metric = data_source_scan.resolve_metric(
            SchemaMetric(
                data_source_scan=self.data_source_scan,
                partition=partition,
                check=self,
            )
        )
        self.metrics[KEY_SCHEMA_MEASURED] = schema_metric

        schema_check_cfg: SchemaCheckCfg = self.check_cfg
        if schema_check_cfg.has_change_validations():
            historic_descriptor = HistoricChangeOverTimeDescriptor(
                metric_identity=schema_metric.identity, change_over_time_cfg=ChangeOverTimeCfg()
            )
            self.historic_descriptors[KEY_SCHEMA_PREVIOUS] = historic_descriptor

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        schema_check_cfg: SchemaCheckCfg = self.check_cfg

        self.measured_schema: list[dict[str, str]] = metrics.get(KEY_SCHEMA_MEASURED).value

        schema_previous_measurement = (
            historic_values.get(KEY_SCHEMA_PREVIOUS).get("measurements").get("results")[0].get("value")
            if historic_values
            and historic_values.get(KEY_SCHEMA_PREVIOUS, {}).get("measurements", {}).get("results", {})
            else None
        )
        schema_previous = (
            [{"name": sp.get("columnName"), "type": sp.get("sourceDataType")} for sp in schema_previous_measurement]
            if schema_previous_measurement
            else None
        )

        if schema_previous:
            self.schema_comparator = SchemaComparator(schema_previous, self.measured_schema)
        else:
            if schema_check_cfg.has_change_validations():
                warning_message = "Skipping schema checks since there is no historic schema metrics!"
                self.logs.warning(warning_message)
                self.add_outcome_reason(outcome_type="notEnoughHistory", message=warning_message, severity="warn")
                return

        self.warn_result = self.get_schema_violations(schema_check_cfg.warn_validations)
        self.fail_result = self.get_schema_violations(schema_check_cfg.fail_validations)

        if self.fail_result:
            self.outcome = CheckOutcome.FAIL
        elif self.warn_result:
            self.outcome = CheckOutcome.WARN
        else:
            self.outcome = CheckOutcome.PASS

    def get_schema_violations(
        self,
        schema_validations: SchemaValidations,
    ) -> SchemaCheckValidationResult | None:
        if schema_validations is None:
            return None

        measured_schema = self.measured_schema

        measured_column_names = [column["name"] for column in measured_schema]

        column_types = {column["name"]: column["type"] for column in measured_schema}

        schema_missing_column_names = []
        schema_present_column_names = []
        schema_column_type_mismatches = {}
        schema_column_index_mismatches = {}

        schema_column_additions = []
        schema_column_deletions = []
        schema_column_type_changes = {}
        schema_column_index_changes = {}

        required_column_names = set()
        if schema_validations.required_column_names:
            required_column_names.update(schema_validations.required_column_names)
        if schema_validations.required_column_types:
            required_column_names.update(schema_validations.required_column_types.keys())

        if required_column_names:
            for required_column_name in required_column_names:
                if required_column_name not in measured_column_names:
                    schema_missing_column_names.append(required_column_name)

        if schema_validations.forbidden_column_names:
            for forbidden_column_name in schema_validations.forbidden_column_names:
                regex = forbidden_column_name.replace("%", ".*").replace("*", ".*")
                forbidden_pattern = re.compile(regex)
                for column_name in measured_column_names:
                    if forbidden_pattern.match(column_name):
                        schema_present_column_names.append(column_name)

        if schema_validations.required_column_types:
            data_source = self.data_source_scan.data_source
            for (
                expected_column_name,
                expected_column_type,
            ) in schema_validations.required_column_types.items():
                if expected_column_name in column_types:
                    actual_type = column_types[expected_column_name]
                    is_same_type = data_source.is_same_type_in_schema_check(expected_column_type, actual_type)
                    if expected_column_name in column_types and not is_same_type:
                        schema_column_type_mismatches[expected_column_name] = {
                            "expected_type": expected_column_type,
                            "actual_type": column_types[expected_column_name],
                        }

        if schema_validations.required_column_indexes:
            for required_column_name in schema_validations.required_column_indexes:
                if required_column_name not in column_types:
                    schema_missing_column_names.append(required_column_name)

            measured_column_indexes = {
                measured_column_name: index for index, measured_column_name in enumerate(measured_column_names)
            }

            for (
                expected_column_name,
                expected_index,
            ) in schema_validations.required_column_indexes.items():
                if expected_column_name in measured_column_indexes:
                    actual_index = measured_column_indexes[expected_column_name]
                    if expected_index != actual_index:
                        column_on_index = (
                            measured_schema[expected_index]["name"] if expected_index < len(measured_schema) else None
                        )
                        schema_column_index_mismatches[expected_column_name] = {
                            "expected_index": expected_index,
                            "actual_index": actual_index,
                            "column_on_expected_index": column_on_index,
                        }
        if self.schema_comparator:
            if schema_validations.is_column_addition_forbidden:
                schema_column_additions = self.schema_comparator.schema_column_additions

            if schema_validations.is_column_deletion_forbidden:
                schema_column_deletions = self.schema_comparator.schema_column_deletions

            if schema_validations.is_column_type_change_forbidden:
                schema_column_type_changes = self.schema_comparator.schema_column_type_changes

            if schema_validations.is_column_index_change_forbidden:
                schema_column_index_changes = self.schema_comparator.schema_column_index_changes

        validation_results = [
            schema_missing_column_names,
            schema_present_column_names,
            schema_column_type_mismatches,
            schema_column_index_mismatches,
            schema_column_additions,
            schema_column_deletions,
            schema_column_type_changes,
            schema_column_index_changes,
        ]

        if any(validation_results):
            return SchemaCheckValidationResult(
                schema_missing_column_names,
                schema_present_column_names,
                schema_column_type_mismatches,
                schema_column_index_mismatches,
                schema_column_additions,
                schema_column_deletions,
                schema_column_type_changes,
                schema_column_index_changes,
            )

        return None

    def get_cloud_diagnostics_dict(self) -> dict:
        schema_diagnostics = {
            "blocks": [],
        }

        if self.measured_schema:
            columns_str = "\n".join([f'{c["name"]},{c["type"]}' for c in self.measured_schema])
            schema_diagnostics["blocks"].append(
                {
                    "type": "csv",
                    "text": f"Column,Type\n{columns_str}",
                    "title": "Schema",
                }
            )

        fail_change_events: list(dict(str, str)) = (
            self.__build_change_events(self.fail_result) if self.fail_result else []
        )
        warn_change_events: list(dict(str, str)) = (
            self.__build_change_events(self.warn_result) if self.warn_result else []
        )

        event_count = len(fail_change_events) + len(warn_change_events)

        if event_count > 0:
            fail_change_events_text = self.__build_change_events_text(fail_change_events, CheckOutcome.FAIL.name)
            warn_change_events_text = self.__build_change_events_text(warn_change_events, CheckOutcome.WARN.name)
            change_events_text = "\n".join([fail_change_events_text, warn_change_events_text])

            schema_diagnostics["blocks"].append(
                {
                    "type": "csv",
                    "text": f"Column,Event,Details\n{change_events_text}",
                    "title": "Diagnostics",
                }
            )

        schema_diagnostics["preferredChart"] = "bars"
        schema_diagnostics["valueLabel"] = f"{event_count} schema event(s)"
        schema_diagnostics["valueSeries"] = {
            "values": [
                {
                    "label": CheckOutcome.FAIL,
                    "value": len(fail_change_events),
                    "outcome": CheckOutcome.FAIL,
                },
                {
                    "label": CheckOutcome.WARN,
                    "value": len(warn_change_events),
                    "outcome": CheckOutcome.WARN,
                },
                {
                    "label": CheckOutcome.PASS,
                    "value": len(self.measured_schema) - event_count,
                    "outcome": CheckOutcome.PASS,
                },
            ]
        }

        return schema_diagnostics

    def __build_change_events(self, schema_validation_result: SchemaCheckValidationResult) -> list(dict(str, str)):
        change_events: list(dict(str, str)) = []

        for column in schema_validation_result.column_deletions:
            change_events.append(
                {
                    "column": column,
                    "event": "Column Deleted",
                    "details": "",
                }
            )
        for column in schema_validation_result.column_additions:
            change_events.append(
                {
                    "column": column,
                    "event": "Column Added",
                    "details": "",
                }
            )
        for column, index_change in schema_validation_result.column_index_changes.items():
            change_events.append(
                {
                    "column": column,
                    "event": "Index Changed",
                    "details": f"Previous Index: {index_change['previous_index']}; New Index: {index_change['new_index']}",
                }
            )
        for column, type_change in schema_validation_result.column_type_changes.items():
            change_events.append(
                {
                    "column": column,
                    "event": "Type Changed",
                    "details": f"Previous Type: {type_change['previous_type']}; New Type: {type_change['new_type']}",
                }
            )

        # Create diagnostics blocks without previous schema comparison as well.
        # Simply take outcome type and count of all changes.
        for column, index_mismatch in schema_validation_result.column_index_mismatches.items():
            change_events.append(
                {
                    "column": column,
                    "event": "Index Mismatch",
                    "details": f"Expected Index: {index_mismatch['expected_index']}; Actual Index: {index_mismatch['actual_index']}",
                }
            )
        for column, type_mismatch in schema_validation_result.column_type_mismatches.items():
            change_events.append(
                {
                    "column": column,
                    "event": "Type Mismatch",
                    "details": f"Expected Type: {type_mismatch['expected_type']}; Actual Type: {type_mismatch['actual_type']}",
                }
            )
        for column in schema_validation_result.present_column_names:
            change_events.append(
                {
                    "column": column,
                    "event": "Forbidden Column Present",
                    "details": "",
                }
            )
        for column in schema_validation_result.missing_column_names:
            change_events.append(
                {
                    "column": column,
                    "event": "Required Column Missing",
                    "details": "",
                }
            )
        return change_events

    def __build_change_events_text(self, change_events: list, event_type: str) -> str:
        return "\n".join(
            f"{ch['column']},:icon-{event_type.lower()}: {ch['event']}, {ch['details']}" for ch in change_events
        )

    def get_log_diagnostic_lines(self) -> list[str]:
        diagnostics_texts: list[str] = []

        if self.outcome in [CheckOutcome.FAIL, CheckOutcome.WARN]:
            if self.fail_result:
                diagnostics_texts.extend(
                    self.__build_log_diagnostic_lines(self.fail_result, CheckOutcome.FAIL.name.lower())
                )
            if self.warn_result:
                diagnostics_texts.extend(
                    self.__build_log_diagnostic_lines(self.warn_result, CheckOutcome.WARN.name.lower())
                )

        diagnostics_texts.append(f"schema_measured = {self.__schema_diagnostics_texts()}")

        return diagnostics_texts

    def __build_log_diagnostic_lines(
        self, schema_validation_result: SchemaCheckValidationResult, type: str
    ) -> list[str]:
        diagnostics_texts: list[str] = []

        if schema_validation_result.missing_column_names:
            diagnostics_texts.append(
                f"{type}_missing_column_names = {self.__list_of_texts(schema_validation_result.missing_column_names)}"
            )
        if schema_validation_result.present_column_names:
            diagnostics_texts.append(
                f"{type}_forbidden_present_column_names = {self.__list_of_texts(schema_validation_result.present_column_names)}"
            )
        if schema_validation_result.column_type_mismatches:
            for (
                column_name,
                column_type_mismatch_data,
            ) in schema_validation_result.column_type_mismatches.items():
                expected_type = column_type_mismatch_data["expected_type"]
                actual_type = column_type_mismatch_data["actual_type"]
                diagnostics_texts.append(
                    f"{type}_column_type_mismatch[{column_name}] expected({expected_type}) actual({actual_type})"
                )
        if schema_validation_result.column_index_mismatches:
            for (
                column_name,
                column_index_mismatch_data,
            ) in schema_validation_result.column_index_mismatches.items():
                expected_index = column_index_mismatch_data["expected_index"]
                actual_index = column_index_mismatch_data["actual_index"]
                column_on_expected_index = column_index_mismatch_data["column_on_expected_index"]
                column_on_expected_index_text = (
                    f"{type}_column_on_expected_index({column_on_expected_index})" if column_on_expected_index else ""
                )
                diagnostics_texts.append(
                    f"{type}_column_index_mismatch[{column_name}] expected({expected_index}) actual({actual_index}){column_on_expected_index_text}"
                )

        if self.schema_comparator:
            if schema_validation_result.column_additions:
                diagnostics_texts.append(
                    f"{type}_column_additions = {self.__list_of_texts(schema_validation_result.column_additions)}"
                )
            if schema_validation_result.column_deletions:
                diagnostics_texts.append(
                    f"{type}_column_deletions = {self.__list_of_texts(schema_validation_result.column_deletions)}"
                )
            if schema_validation_result.column_type_changes:
                for (
                    column_name,
                    column_type_change_data,
                ) in schema_validation_result.column_type_changes.items():
                    previous_type = column_type_change_data["previous_type"]
                    new_type = column_type_change_data["new_type"]
                    diagnostics_texts.append(
                        f"{type}_column_type_change[{column_name}] new_type({new_type}) previous_type({previous_type})"
                    )
            if schema_validation_result.column_index_changes:
                changes = []
                for (
                    column_name,
                    column_index_change_data,
                ) in schema_validation_result.column_index_changes.items():
                    previous_index = column_index_change_data["previous_index"]
                    new_index = column_index_change_data["new_index"]
                    changes.append(f"{column_name}[{previous_index}->{new_index}]")
                changes_txt = ", ".join(changes)
                diagnostics_texts.append(f"{type}_column_index_changes = {changes_txt}")

        return diagnostics_texts

    def __list_of_texts(self, texts):
        elements = ", ".join(texts)
        return f"[{elements}]"

    def __schema_diagnostics_texts(self):
        return self.__list_of_texts((column["name"] + " " + column["type"]) for column in self.measured_schema)



from __future__ import annotations


class Table:
    def __init__(self, data_source_scan: DataSourceScan, table_name: str):
        from soda.execution.column import Column
        from soda.execution.partition import Partition

        self.data_source_scan = data_source_scan
        self.table_name = str(table_name)
        self.qualified_table_name = self.data_source_scan.data_source.qualified_table_name(table_name)
        self.columns_by_name: dict[str, Column] = {}
        self.partitions: dict[str, Partition] = {}

    def get_or_create_column(self, column_name: str):
        column = self.columns_by_name.get(column_name)
        if not column:
            from soda.execution.column import Column

            column = Column(self, column_name)
            self.columns_by_name[column_name] = column
        return column

    def get_or_create_partition(self, partition_name: str | None):
        partition = self.partitions.get(partition_name)
        if partition is None:
            from soda.execution.partition import Partition

            partition = Partition(self, partition_name)
            self.partitions[partition_name] = partition
        return partition



from typing import Dict, List

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.data_source_check_cfg import DataSourceCheckCfg
from soda.sodacl.table_cfg import TableCfg


class DataSourceScanCfg:
    def __init__(self, data_source_name: str):
        self.data_source_name: str = data_source_name
        self.tables_cfgs: Dict[str, TableCfg] = {}
        self.data_source_cfgs: List[DataSourceCheckCfg] = []
        self.check_cfgs: List[CheckCfg] = []

    def get_or_create_table_cfg(self, table_name) -> TableCfg:
        table_cfg = self.tables_cfgs.get(table_name)
        if not table_cfg:
            table_cfg = TableCfg(table_name)
            self.tables_cfgs[table_name] = table_cfg
        return table_cfg

    def add_data_source_cfg(self, data_source_check_cfg: DataSourceCheckCfg):
        self.data_source_cfgs.append(data_source_check_cfg)

    def add_check_cfg(self, check_cfg: CheckCfg):
        self.check_cfgs.append(check_cfg)



import logging
from typing import Dict, Optional, Sequence

from opentelemetry.exporter.otlp.proto.http import Compression
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SpanExportResult

logger = logging.getLogger(__name__)


def get_soda_spans(spans: Sequence[ReadableSpan]) -> Sequence[ReadableSpan]:
    result = []
    for span in spans:
        if span.name.startswith("soda"):
            result.append(span)
        else:
            logger.debug(f"Open Telemetry: Skipping non-soda span '{span.name}'.")

    return result


class SodaConsoleSpanExporter(ConsoleSpanExporter):
    """Soda version of console exporter.

    Does not export any non-soda spans for security and privacy reasons."""

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        return super().export(get_soda_spans(spans))


class SodaOTLPSpanExporter(OTLPSpanExporter):
    """Soda version of OTLP exporter.

    Does not export any non-soda spans for security and privacy reasons."""

    def __init__(
        self,
        endpoint: Optional[str] = None,
        certificate_file: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[int] = None,
        compression: Optional[Compression] = None,
    ):
        super().__init__(
            endpoint,
            certificate_file,
            headers,
            timeout,
            compression,
        )

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        return super().export(get_soda_spans(spans))



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from soda.profiling.numeric_column_profiler import NumericColumnProfiler
from soda.profiling.profile_columns_result import (
    ProfileColumnsResult,
    ProfileColumnsResultColumn,
    ProfileColumnsResultTable,
)
from soda.profiling.text_column_profiler import TextColumnProfiler

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan
    from soda.sodacl.data_source_check_cfg import ProfileColumnsCfg


class ProfileColumnsRun:
    def __init__(self, data_source_scan: DataSourceScan, profile_columns_cfg: ProfileColumnsCfg):
        self.data_source_scan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.profile_columns_cfg: ProfileColumnsCfg = profile_columns_cfg
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> ProfileColumnsResult:
        self.logs.info(f"Running column profiling for data source: {self.data_source.data_source_name}")

        profile_result_column_tables = self.get_table_columns_metadata()

        profile_columns_result: ProfileColumnsResult = ProfileColumnsResult(self.profile_columns_cfg)

        if profile_result_column_tables is None:
            self._raise_no_tables_to_profile_warning()
            return profile_columns_result

        self.logs.info("Profiling columns for the following tables:")
        for table_name, columns_metadata in profile_result_column_tables.items():
            row_count = self.data_source.get_table_row_count(table_name)
            result_table = ProfileColumnsResultTable(
                table_name=table_name, data_source=self.data_source.data_source_name, row_count=row_count
            )
            self.logs.info(f"  - {table_name}")
            for column_name, column_data_type in columns_metadata.items():
                profiling_column_type = "unknown type"
                try:
                    if column_data_type.startswith(tuple(self.data_source.NUMERIC_TYPES_FOR_PROFILING)):
                        profiling_column_type = "numeric"
                        numeric_column_profiler = NumericColumnProfiler(
                            data_source_scan=self.data_source_scan,
                            profile_columns_cfg=self.profile_columns_cfg,
                            table_name=table_name,
                            column_name=column_name,
                            column_data_type=column_data_type,
                        )
                        result_column: ProfileColumnsResultColumn = numeric_column_profiler.profile()
                        result_table.append_column(result_column)
                    elif column_data_type.startswith(tuple(self.data_source.TEXT_TYPES_FOR_PROFILING)):
                        profiling_column_type = "text"
                        text_column_profiler = TextColumnProfiler(
                            data_source_scan=self.data_source_scan,
                            profile_columns_cfg=self.profile_columns_cfg,
                            table_name=table_name,
                            column_name=column_name,
                            column_data_type=column_data_type,
                        )
                        result_column: ProfileColumnsResultColumn = text_column_profiler.profile()
                        result_table.append_column(result_column)
                    else:
                        self.logs.warning(
                            f"Column '{table_name}.{column_name}' was not profiled because column data "
                            f"type '{column_data_type}' is not in supported profiling data types"
                        )
                except Exception as e:
                    self.logs.error(
                        f"Problem profiling {profiling_column_type} column '{table_name}.{column_name}' with data type '{column_data_type}': {e}"
                    )
            profile_columns_result.append_table(result_table)
        return profile_columns_result

    def get_table_columns_metadata(self) -> defaultdict[str, dict] | None:
        include_patterns = self.parse_profiling_expressions(self.profile_columns_cfg.include_columns)
        exclude_patterns = self.parse_profiling_expressions(self.profile_columns_cfg.exclude_columns)

        tables_columns_metadata = self.data_source.get_tables_columns_metadata(
            include_patterns=include_patterns,
            exclude_patterns=exclude_patterns,
            query_name="profile-columns-get-table-and-column-metadata",
        )
        if not tables_columns_metadata:
            return None
        return tables_columns_metadata

    @staticmethod
    def parse_profiling_expressions(profiling_expressions: list[str]) -> list[dict[str, str]]:
        parsed_profiling_expressions = []
        for profiling_expression in profiling_expressions:
            table_name_pattern, column_name_pattern = profiling_expression.split(".")
            parsed_profiling_expressions.append(
                {
                    "table_name_pattern": table_name_pattern,
                    "column_name_pattern": column_name_pattern,
                }
            )
        return parsed_profiling_expressions

    def _raise_no_tables_to_profile_warning(self) -> None:
        self.logs.warning(
            "Your SodaCL profiling expressions did not return any existing dataset name"
            f" and column name combinations for your '{self.data_source.data_source_name}' "
            "data source. Please make sure that the patterns in your profiling expressions define "
            "existing dataset name and column name combinations."
            " Profiling results may be incomplete or entirely skipped. See the docs for more information: \n"
            f"https://go.soda.io/display-profile",
            location=self.profile_columns_cfg.location,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from soda.profiling.discover_tables_result import DiscoverTablesResult
from soda.sodacl.data_source_check_cfg import DataSourceCheckCfg

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan


class DiscoverTablesRun:
    def __init__(self, data_source_scan: DataSourceScan, data_source_check_cfg: DataSourceCheckCfg):
        self.data_source_scan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.data_source_check_cfg: DataSourceCheckCfg = data_source_check_cfg
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> DiscoverTablesResult:
        discover_tables_result: DiscoverTablesResult = DiscoverTablesResult(self.data_source_check_cfg)
        self.logs.info(f"Running discover datasets for data source: {self.data_source.data_source_name}")

        # row_counts is a dict that maps table names to row counts.
        table_names: dict[str, int] = self.data_source.get_table_names(
            include_tables=self.data_source_check_cfg.include_tables,
            exclude_tables=self.data_source_check_cfg.exclude_tables,
            query_name="discover-tables-find-tables-and-row-counts",
        )

        if len(table_names) < 1:
            self.logs.warning(
                f"No table matching your SodaCL inclusion list found on your {self.data_source.data_source_name} "
                "data source. Table discovery results may be incomplete or entirely skipped",
                location=self.data_source_check_cfg.location,
            )
            return discover_tables_result

        self.logs.info(f"Discovering the following tables:")
        for table_name in table_names:
            self.logs.info(f"  - {table_name}")
            discover_tables_result_table = discover_tables_result.create_table(
                table_name, self.data_source.data_source_name
            )
            # get columns & metadata for current table
            columns_metadata_result = self.data_source.get_table_columns(
                table_name=table_name,
                query_name=f"discover-tables-column-metadata-for-{table_name}",
            )

            for column_name, column_type in columns_metadata_result.items():
                _ = discover_tables_result_table.create_column(column_name, column_type)

        return discover_tables_result



from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from textwrap import indent

from soda.common.exception_helper import get_exception_stacktrace
from soda.sodacl.location import Location

logger = logging.getLogger("soda.scan")


class LogLevel(Enum):
    WARNING = "WARNING"
    ERROR = "ERROR"
    INFO = "INFO"
    DEBUG = "DEBUG"


@dataclass
class Log:
    __log_level_mappings = {
        LogLevel.ERROR: logging.ERROR,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.INFO: logging.INFO,
        LogLevel.DEBUG: logging.DEBUG,
    }
    __soda_cloud_level_mappings = {
        LogLevel.ERROR: "error",
        LogLevel.WARNING: "warning",
        LogLevel.INFO: "info",
        LogLevel.DEBUG: "debug",
    }

    def __init__(
        self,
        level: LogLevel,
        message: str,
        location: Location | None,
        doc: str | None,
        exception: BaseException | None,
        timestamp: datetime | None = None,
    ):
        self.level: LogLevel = level
        self.message: str = message
        self.location: Location | None = location
        self.doc: str | None = doc
        self.exception: BaseException | None = exception
        self.timestamp: datetime = timestamp if isinstance(timestamp, datetime) else datetime.now(tz=timezone.utc)
        self.index = self.get_next_index()

    __index = 0

    @classmethod
    def get_next_index(cls):
        cls.__index += 1
        return cls.__index

    def __str__(self):
        location_str = f" | {self.location}" if self.location else ""
        doc_str = f" | https://go.soda.io/{self.doc}" if self.doc else ""
        exception_str = f" | {self.exception}" if self.exception else ""
        return f"{self.level.value.ljust(7)}| {self.message}{location_str}{doc_str}{exception_str}"

    def get_cloud_dict(self) -> dict:
        log_cloud_dict = {
            "level": self.__soda_cloud_level_mappings[self.level],
            "message": self.message,
            "timestamp": self.timestamp,
            "index": self.index,
        }
        if self.location:
            location_cloud_dict = self.location.get_cloud_dict()
            log_cloud_dict["errorLocation"] = location_cloud_dict
            log_cloud_dict["location"] = location_cloud_dict
        if self.doc:
            log_cloud_dict["doc"] = self.doc
        return log_cloud_dict

    def get_dict(self) -> dict:
        return {
            "level": self.level,
            "message": self.message,
            "timestamp": self.timestamp,
            "index": self.index,
            "doc": self.doc if self.doc else None,
            "location": self.location.get_dict() if self.location else None,
        }

    @staticmethod
    def log_errors(error_logs):
        logger.info(f"ERRORS:")
        for error_log in error_logs:
            error_log.log_to_python_logging()

    def log_to_python_logging(self):
        python_log_level = Log.__log_level_mappings[self.level]
        time = datetime.now().strftime("%H:%M:%S")
        logger.log(python_log_level, f"[{time}] {self.message}")
        if self.exception is not None:
            exception_str = str(self.exception)
            exception_str = indent(text=exception_str, prefix="  | ")
            logger.log(python_log_level, exception_str)
            # Logging the stack trace
            from soda.scan import verbose

            if verbose:
                stacktrace = get_exception_stacktrace(self.exception)
                indented_stacktrace = indent(text=stacktrace, prefix="  | ")
                logger.log(python_log_level, f"  | Stacktrace:")
                logger.log(python_log_level, indented_stacktrace)
        if self.level in [LogLevel.WARNING, LogLevel.ERROR] and self.location is not None:
            logger.log(python_log_level, f"  +-> {self.location}")
        if isinstance(self.doc, str):
            logger.log(python_log_level, f"  +-> See https://go.soda.io/{self.doc}")



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class SampleColumn:
    name: str
    type: str

    def get_cloud_dict(self):
        return {"name": self.name, "type": self.type}

    def get_dict(self):
        return {"name": self.name, "type": self.type}

    @staticmethod
    def create_sample_columns(dbapi_description, data_source) -> list[SampleColumn]:
        return [
            SampleColumn._convert_python_db_column_to_sample_column(dbapi_column, data_source)
            for dbapi_column in dbapi_description
        ]

    @staticmethod
    def _convert_python_db_column_to_sample_column(dbapi_column, data_source):
        type_code = dbapi_column[1]
        type_name = data_source.get_type_name(type_code)
        return SampleColumn(name=dbapi_column[0], type=type_name)


@dataclass
class SampleSchema:
    columns: list[SampleColumn]

    def get_dict(self):
        return [x.get_dict() for x in self.columns]



from __future__ import annotations

import re

import sqlparse


def parse_columns_from_query(query: str) -> list[str]:
    # Handle non-standard sql syntax caveats before parsing.
    # Remove sqlserver TOP from select statement.
    query = re.sub(r"(?i)top(\(.*\)|\s+\d+)", "", query)

    columns = []
    statements = sqlparse.split(query)

    for statement in statements:
        column_lines = None
        start_looking_for_columns = False
        parsed = sqlparse.parse(statement)[0]
        for token in parsed.tokens:
            if token.ttype == sqlparse.tokens.Keyword.DML and token.value.lower() == "select":
                start_looking_for_columns = True

            # This is now after a "select", look for a) no type which is list of columns or b) wildcard which is "*"
            if start_looking_for_columns and (token.ttype == None or token.ttype == sqlparse.tokens.Token.Wildcard):
                column_lines = token.value
                break
        # Remove newlines, double whitespaces and split by colon.
        column_lines = " ".join(column_lines.replace("\n", "").split()).split(",")

        # Strip each "column statement" of whitespace, remove everything after the first whitespace and keep everything after first dot.
        for column in column_lines:
            column, _, _ = column.strip().partition(" ")
            if "." in column:
                _, _, column = column.strip().partition(".")
            columns.append(column)

    return columns



from __future__ import annotations

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


class UserDefinedFailedRowsExpressionCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: dict | None,
        location: Location,
        name: str,
        fail_condition_sql_expr: str | None,
        samples_limit: int | None = None,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.samples_limit = samples_limit
        self.fail_condition_sql_expr: str | None = fail_condition_sql_expr



from __future__ import annotations

from typing import TYPE_CHECKING

from soda.profiling.sample_tables_result import SampleTablesResult
from soda.sodacl.data_source_check_cfg import DataSourceCheckCfg

if TYPE_CHECKING:
    from soda.execution.data_source_scan import DataSourceScan


class SampleTablesRun:
    def __init__(self, data_source_scan: DataSourceScan, data_source_check_cfg: DataSourceCheckCfg):
        self.data_source_scan = data_source_scan
        self.soda_cloud = data_source_scan.scan._configuration.soda_cloud
        self.data_source = data_source_scan.data_source
        self.data_source_name = self.data_source.data_source_name
        self.data_source_check_cfg: DataSourceCheckCfg = data_source_check_cfg
        self.logs = self.data_source_scan.scan._logs

    def run(self) -> SampleTablesResult:
        sample_tables_result: SampleTablesResult = SampleTablesResult(self.data_source_check_cfg)
        self.logs.info(f"Running sample tables for data source: {self.data_source_name}")

        table_names: list[str] = self.data_source.get_table_names(
            include_tables=self.data_source_check_cfg.include_tables,
            exclude_tables=self.data_source_check_cfg.exclude_tables,
            query_name=f"sample-datasets-find-datasets",
        )

        self.logs.info(f"Sampling the following tables:")
        for table_name in table_names:
            self.logs.info(f"  - {table_name}")

            # get columns and first n rows
            sample_ref = self.data_source.store_table_sample(table_name, limit=100)
            sample_tables_result.append_table(table_name, self.data_source_name, sample_ref)
            self.logs.info(f"Successfully collected samples for dataset: {table_name}!")

        if not isinstance(sample_tables_result.tables, list):
            self.logs.error(f"Sample tables for data source: {self.data_source_name} failed")

        return sample_tables_result



from __future__ import annotations

import json
import logging
import re
import tempfile
from datetime import date, datetime, timedelta, timezone
from typing import TYPE_CHECKING

import requests
from requests import Response
from soda.__version__ import SODA_CORE_VERSION
from soda.common.json_helper import JsonHelper
from soda.common.logs import Logs
from soda.execution.check_type import CheckType
from soda.soda_cloud.historic_descriptor import (
    HistoricChangeOverTimeDescriptor,
    HistoricCheckResultsDescriptor,
    HistoricDescriptor,
    HistoricMeasurementsDescriptor,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from soda.scan import Scan

GENERIC_TYPE_CSV_TEXT_MAX_LENGTH = 1500


class SodaCloud:
    ORG_CONFIG_KEY_CHECK_ATTRIBUTES = "sodaCoreAvailableCheckAttributes"
    ORG_CONFIG_KEY_DISABLE_COLLECTING_WH_DATA = "disableCollectingWarehouseData"

    def __init__(
        self,
        host: str,
        api_key_id: str,
        api_key_secret: str,
        token: str | None,
        port: str | None,
        logs: Logs,
        scheme: str = "https",
    ):
        self.host = host
        self.port = f":{port}" if port else ""
        self.scheme = scheme if scheme else "https"
        self.api_url = f"{self.scheme}://{self.host}{self.port}/api"
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.token: str | None = token
        self.headers = {"User-Agent": f"SodaCore/{SODA_CORE_VERSION}"}
        self.logs = logs
        self.soda_cloud_trace_ids = {}
        self._organization_configuration = None

    @property
    def organization_configuration(self) -> dict:
        if isinstance(self._organization_configuration, dict):
            return self._organization_configuration

        response_json_dict = self._execute_query(
            {"type": "sodaCoreCloudConfiguration"},
            query_name="get_organization_configuration",
        )
        self._organization_configuration = response_json_dict if isinstance(response_json_dict, dict) else {}

        return self._organization_configuration

    @staticmethod
    def build_scan_results(scan) -> dict:
        checks = [
            check.get_cloud_dict()
            for check in scan._checks
            if check.check_type == CheckType.CLOUD
            and (check.outcome is not None or check.force_send_results_to_cloud is True)
            and check.archetype is None
        ]
        automated_monitoring_checks = [
            check.get_cloud_dict()
            for check in scan._checks
            if (check.outcome is not None or check.force_send_results_to_cloud is True) and check.archetype is not None
        ]

        # TODO: [SODA-608] separate profile columns and sample tables by aligning with the backend team
        profiling = [
            profile_table.get_cloud_dict()
            for profile_table in scan._profile_columns_result_tables + scan._sample_tables_result_tables
        ]

        query_list = []
        for query in scan._queries:
            query_list += query.get_cloud_dicts()

        return JsonHelper.to_jsonnable(  # type: ignore
            {
                "definitionName": scan._scan_definition_name,
                "defaultDataSource": scan._data_source_name,
                "dataTimestamp": scan._data_timestamp,
                # Can be changed by user, this is shown in Cloud as time of a scan.
                "scanStartTimestamp": scan._scan_start_timestamp,  # Actual time when the scan started.
                "scanEndTimestamp": scan._scan_end_timestamp,  # Actual time when scan ended.
                "hasErrors": scan.has_error_logs(),
                "hasWarnings": scan.has_check_warns(),
                "hasFailures": scan.has_check_fails(),
                "metrics": [metric.get_cloud_dict() for metric in scan._metrics],
                # If archetype is not None, it means that check is automated monitoring
                "checks": checks,
                "queries": query_list,
                "automatedMonitoringChecks": automated_monitoring_checks,
                "profiling": profiling,
                "metadata": [
                    discover_tables_result.get_cloud_dict()
                    for discover_tables_result in scan._discover_tables_result_tables
                ],
                "logs": [log.get_cloud_dict() for log in scan._logs.logs],
            }
        )

    @staticmethod
    def _serialize_file_upload_value(value):
        if value is None or isinstance(value, str) or isinstance(value, int) or isinstance(value, float):
            return value
        return str(value)

    def upload_sample(
        self, scan: Scan, sample_rows: tuple[tuple], sample_file_name: str, samples_limit: int | None
    ) -> str:
        """
        :param sample_file_name: file name without extension
        :return: Soda Cloud file_id
        """

        # Keep the interface of this method backward compatible and allow for samples limit to be None, but do not continue with no limit in such case.
        if not samples_limit:
            samples_limit = 100

        try:
            scan_definition_name = scan._scan_definition_name
            scan_data_timestamp = scan._data_timestamp
            scan_folder_name = (
                f"{self._fileify(scan_definition_name)}"
                f'_{scan_data_timestamp.strftime("%Y%m%d%H%M%S")}'
                f'_{datetime.now(tz=timezone.utc).strftime("%Y%m%d%H%M%S")}'
            )

            with tempfile.TemporaryFile() as temp_file:
                for row in sample_rows[0:samples_limit]:
                    row = [self._serialize_file_upload_value(v) for v in row]
                    rows_json_str = json.dumps(row)
                    rows_json_bytes = bytearray(rows_json_str, "utf-8")
                    temp_file.write(rows_json_bytes)
                    temp_file.write(b"\n")

                temp_file_size_in_bytes = temp_file.tell()
                temp_file.seek(0)

                file_path = f"{scan_folder_name}/" + f"{sample_file_name}.jsonl"

                file_id = self._upload_sample_http(scan_definition_name, file_path, temp_file, temp_file_size_in_bytes)

                return file_id

        except Exception as e:
            self.logs.error(f"Soda cloud error: Could not upload sample {sample_file_name}", exception=e)

    def _fileify(self, name: str):
        return re.sub(r"\W+", "_", name).lower()

    def _upload_sample_http(self, scan_definition_name: str, file_path, temp_file, file_size_in_bytes: int):
        headers = {
            "Authorization": self._get_token(),
            "Content-Type": "application/octet-stream",
            "Is-V3": "true",
            "File-Path": file_path,
        }

        if file_size_in_bytes == 0:
            # because of https://github.com/psf/requests/issues/4215 we can't send content size
            # when the size is 0 since requests blocks then on I/O indefinitely
            self.logs.warning("Empty file upload detected, not sending Content-Length header")
        else:
            headers["Content-Length"] = str(file_size_in_bytes)

        upload_response = self._http_post(url=f"{self.api_url}/scan/upload", headers=headers, data=temp_file)
        upload_response_json = upload_response.json()

        if "fileId" not in upload_response_json:
            logger.error(f"No fileId received in response: {upload_response_json}")
            return None
        else:
            return upload_response_json["fileId"]

    def send_scan_results(self, scan: Scan):
        scan_results = self.build_scan_results(scan)
        scan_results["type"] = "sodaCoreInsertScanResults"
        return self._execute_command(scan_results, command_name="send_scan_results")

    def get_historic_data(self, historic_descriptor: HistoricDescriptor):
        measurements = {}
        check_results = {}

        if type(historic_descriptor) == HistoricMeasurementsDescriptor:
            measurements = self._get_historic_measurements(historic_descriptor)
        elif type(historic_descriptor) == HistoricCheckResultsDescriptor:
            check_results = self._get_historic_check_results(historic_descriptor)
        elif type(historic_descriptor) == HistoricChangeOverTimeDescriptor:
            measurements = self._get_historic_changes_over_time(historic_descriptor)
        else:
            logger.error(f"Invalid Historic Descriptor provided {historic_descriptor}")

        return {"measurements": measurements, "check_results": check_results}

    def is_samples_disabled(self) -> bool:
        return self.organization_configuration.get(self.ORG_CONFIG_KEY_DISABLE_COLLECTING_WH_DATA, True)

    def get_check_attributes_schema(self) -> list(dict):
        response_json_dict = self._execute_query(
            {"type": "sodaCoreAvailableCheckAttributes"},
            query_name="get_check_attributes",
        )

        if response_json_dict and "results" in response_json_dict:
            return response_json_dict["results"]

        return []

    def _get_historic_changes_over_time(self, hd: HistoricChangeOverTimeDescriptor):
        query = {
            "type": "sodaCoreHistoricMeasurements",
            "filter": {
                "type": "and",
                "andExpressions": [
                    {
                        "type": "equals",
                        "left": {"type": "columnValue", "columnName": "metric.identity"},
                        "right": {"type": "string", "value": hd.metric_identity},
                    }
                ],
            },
        }

        previous_time_start = None
        previous_time_end = None
        today = date.today()

        if hd.change_over_time_cfg.same_day_last_week:
            last_week = today - timedelta(days=7)
            previous_time_start = datetime(
                year=last_week.year, month=last_week.month, day=last_week.day, tzinfo=timezone.utc
            )
            previous_time_end = datetime(
                year=last_week.year,
                month=last_week.month,
                day=last_week.day,
                hour=23,
                minute=59,
                second=59,
                tzinfo=timezone.utc,
            )

        if previous_time_start and previous_time_end:
            query["filter"]["andExpressions"].append(
                {
                    "type": "greaterThanOrEqual",
                    "left": {"type": "columnValue", "columnName": "measurement.dataTime"},
                    "right": {"type": "time", "scanTime": False, "time": previous_time_start.isoformat()},
                }
            )
            query["filter"]["andExpressions"].append(
                {
                    "type": "lessThanOrEqual",
                    "left": {"type": "columnValue", "columnName": "measurement.dataTime"},
                    "right": {"type": "time", "scanTime": False, "time": previous_time_end.isoformat()},
                }
            )

        return self._execute_query(
            query,
            query_name="get_hisoric_changes_over_time",
        )

    def _get_historic_measurements(self, hd: HistoricMeasurementsDescriptor):
        historic_measurements = self._execute_query(
            {
                "type": "sodaCoreHistoricMeasurements",
                "limit": hd.limit,
                "filter": {
                    "type": "and",
                    "andExpressions": [
                        {
                            "type": "equals",
                            "left": {"type": "columnValue", "columnName": "metric.identity"},
                            "right": {"type": "string", "value": hd.metric_identity},
                        }
                    ],
                },
            },
            query_name="get_hisotric_check_results",
        )
        # Filter out historic_measurements not having 'value' key
        historic_measurements["results"] = [
            measurement for measurement in historic_measurements["results"] if "value" in measurement
        ]
        return historic_measurements

    def _get_historic_check_results(self, hd: HistoricCheckResultsDescriptor):
        return self._execute_query(
            {
                "type": "sodaCoreHistoricCheckResults",
                "limit": hd.limit,
                "filter": {
                    "type": "and",
                    "andExpressions": [
                        {
                            "type": "equals",
                            "left": {"type": "columnValue", "columnName": "check.identity"},
                            "right": {"type": "string", "value": hd.check_identity},
                        }
                    ],
                },
            },
            query_name="get_hisotric_check_results",
        )

    def _execute_query(self, query: dict, query_name: str):
        return self._execute_request("query", query, False, query_name)

    def _execute_command(self, command: dict, command_name: str):
        return self._execute_request("command", command, False, command_name)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool, request_name: str):
        from soda.scan import verbose

        try:
            request_body["token"] = self._get_token()
            if verbose:
                logger.debug(f"{JsonHelper.to_json_pretty(request_body)}")
            response = self._http_post(
                url=f"{self.api_url}/{request_type}", headers=self.headers, json=request_body, request_name=request_name
            )
            response_json = response.json()
            if response.status_code == 401 and not is_retry:
                logger.debug("Authentication failed. Probably token expired. Re-authenticating...")
                self.token = None
                response_json = self._execute_request(request_type, request_body, True, request_name)
            elif response.status_code != 200:
                self.logs.error(
                    f"Error while executing Soda Cloud {request_type} response code: {response.status_code}"
                )
                if verbose:
                    self.logs.debug(response.text)
            return response_json
        except Exception as e:
            self.logs.error(f"Error while executing Soda Cloud {request_type}", exception=e)

    def _get_token(self):
        if not self.token:
            login_command = {"type": "login"}
            if self.api_key_id and self.api_key_secret:
                login_command["apiKeyId"] = self.api_key_id
                login_command["apiKeySecret"] = self.api_key_secret
            else:
                raise RuntimeError("No API KEY and/or SECRET provided ")

            login_response = self._http_post(
                url=f"{self.api_url}/command", headers=self.headers, json=login_command, request_name="get_token"
            )
            if login_response.status_code != 200:
                raise AssertionError(f"Soda Cloud login failed {login_response.status_code}. Check credentials.")
            login_response_json = login_response.json()

            self.token = login_response_json.get("token")
            assert self.token, "No token in login response?!"
        return self.token

    def _http_post(self, request_name: str = None, **kwargs) -> Response:
        response = requests.post(**kwargs)

        if request_name:
            trace_id = response.headers.get("X-Soda-Trace-Id")
            if trace_id:
                self.soda_cloud_trace_ids[request_name] = trace_id

        return response



from dataclasses import dataclass
from typing import List, Optional

from soda.sodacl.check_cfg import CheckCfg
from soda.sodacl.location import Location


@dataclass
class GroupValidations:
    required_group_names: Optional[List[str]]
    forbidden_group_names: Optional[List[str]]
    is_group_addition_forbidden: bool
    is_group_deletion_forbidden: bool

    def has_change_validations(self):
        return self.is_group_addition_forbidden or self.is_group_deletion_forbidden


class GroupEvolutionCheckCfg(CheckCfg):
    def __init__(
        self,
        source_header: str,
        source_line: str,
        source_configurations: Optional[str],
        location: Location,
        name: Optional[str],
        query: str,
        warn_validations: GroupValidations,
        fail_validations: GroupValidations,
    ):
        super().__init__(source_header, source_line, source_configurations, location, name)
        self.warn_validations: GroupValidations = warn_validations
        self.fail_validations: GroupValidations = fail_validations
        self.name = name if name else "Group Evolution Check"
        self.query = query

    def has_change_validations(self) -> bool:
        return (self.warn_validations and self.warn_validations.has_change_validations()) or (
            self.fail_validations and self.fail_validations.has_change_validations()
        )
